# cursor.py
# Copyright 2017 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Define the cursor interface using method names taken from Berkeley DB to
indicate preference for that style of cursor.

Subclasses will provided appropriate implementations for the record definition
classes to use.

"""


class CursorError(Exception):
    pass


class Cursor(object):
    """Define cursor access methods that subclasses must provide.

    These methods will be implemented using the cursor methods of the
    underlying database engine.  Subclasses may also provide methods
    with names matching those of the bsddb interface (typically first
    corresponding to first and so on).

    """

    def __init__(self, dbset):
        """Define a cursor on the underlying database engine."""
        super(Cursor, self).__init__()
        self._cursor = None
        self._dbset = dbset
        self._partial = None

    def close(self):
        """return None if implemented."""
        raise DatabaseError('close not implemented')

    def __del__(self):
        """Call the instance close() method."""
        self.close()

    def count_records(self):
        """return record count or None if implemented."""
        raise DatabaseError('count_records not implemented')

    def database_cursor_exists(self):
        """return True if cursor exists or False if implemented."""
        raise DatabaseError('database_cursor_exists not implemented')

    def first(self):
        """return (key, value) or None if implemented."""
        raise DatabaseError('first not implemented')

    def get_position_of_record(self, record=None):
        """return position of record in file or 0 (zero) if implemented."""
        raise DatabaseError('get_position_of_record not implemented')

    def get_record_at_position(self, position=None):
        """return record for positionth record in file or None if implemented.
        """
        raise DatabaseError('get_record_at_position not implemented')

    def last(self):
        """return (key, value) or None if implemented."""
        raise DatabaseError('last not implemented')

    def nearest(self, key):
        """return (key, value) or None if implemented."""
        raise DatabaseError('nearest not implemented')

    def next(self):
        """return (key, value) or None if implemented."""
        raise DatabaseError('next not implemented')

    def prev(self):
        """return (key, value) or None if implemented."""
        raise DatabaseError('prev not implemented')

    def refresh_recordset(self, instance=None):
        """Amend data structures after database update and return None if
        implemented.

        It may be correct to do nothing.

        """
        raise DatabaseError('refresh_recordset not implemented')

    def setat(self, record):
        """return (key, value) or None if implemented."""
        raise DatabaseError('setat not implemented')

    def set_partial_key(self, partial):
        """return (key, value) or None if implemented."""
        raise DatabaseError('set_partial_key not implemented')

    def get_partial(self):
        """return self._partial if implemented."""
        raise DatabaseError('get_partial not implemented')

    def get_converted_partial(self):
        """return self._partial as it would be held on database if implemented.
        """
        raise DatabaseError('get_converted_partial not implemented')

    def get_partial_with_wildcard(self):
        """return self._partial with wildcard suffix appended if implemented."""
        raise DatabaseError('get_partial_with_wildcard not implemented')

    def get_converted_partial_with_wildcard(self):
        """return converted self._partial with wildcard suffix appended if
        implemented."""
        raise DatabaseError(
            'get_converted_partial_with_wildcard not implemented')
