# sqlite3api_test.py
# Copyright 2014 Roger Marsh
# License: See LICENSE.TXT (BSD licence)

"""sqlite3api tests"""

import unittest
import os
import shutil
import sqlite3
import subprocess
import sys

from ..api.test._filespec import (
    samplefilespec,
    SampleValue,
    SampleRecord,
    GAMES_FILE_DEF,
    GAME_FIELD_DEF,
    EVENT_FIELD_DEF,
    SITE_FIELD_DEF,
    DATE_FIELD_DEF,
    ROUND_FIELD_DEF,
    WHITE_FIELD_DEF,
    BLACK_FIELD_DEF,
    RESULT_FIELD_DEF,
    SEVEN_TAG_ROSTER,
    PRIMARY,
    SECONDARY,
    )
from .. import sqlite3api, _sqlite
from ..api import filespec, record, recordset, database
from ..api.segmentsize import SegmentSize
from . import database_interface


def base(folder):
    return sqlite3api.Sqlite3base(
        sqlite3api.Sqlite3Primary,
        sqlite3api.Sqlite3Secondary,
        samplefilespec(),
        os.path.expanduser(
            os.path.join('~', 'sqlite3api_tests', folder)))


def api(folder):
    return sqlite3api.Sqlite3api(
        samplefilespec(),
        os.path.expanduser(
            os.path.join('~', 'sqlite3api_tests', folder)))


def _delete_folder(folder):
    shutil.rmtree(
        os.path.expanduser(os.path.join('~', 'sqlite3api_tests', folder)),
        ignore_errors=True)


# Defined at module level for pickling.
class _Value(record.Value):
    def pack(self):
        v = super().pack()
        v[1]['Site'] = 'gash' # defined in samplefilespec()
        v[1]['hhhhhh'] = 'hhhh' # not defined in samplefilespec(), ignored
        return v


# Defined at module level for pickling.
class _ValueEdited(record.Value):
    def pack(self):
        v = super().pack()
        v[1]['Site'] = 'newgash' # defined in samplefilespec()
        v[1]['hhhhhh'] = 'hhhh' # not defined in samplefilespec(), ignored
        return v


class _DatabaseEncoders(unittest.TestCase):

    def setUp(self):
        self.sde = _sqlite._DatabaseEncoders()

    def tearDown(self):
        pass

    def test____assumptions(self):
        self.assertEqual(repr(57), '57')

    def test_encode_record_number_01(self):
        self.assertEqual(self.sde.encode_record_number(57), '57')

    def test_decode_record_number_01(self):
        self.assertRaisesRegex(
            ValueError,
            "malformed node or string: 57",
            self.sde.decode_record_number,
            57)
        self.assertEqual(self.sde.decode_record_number('57'), 57)

    def test_encode_record_selector(self):
        self.assertEqual(self.sde.encode_record_selector('57'), '57')


class Sqlite3base_init(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test___init___01(self):
        p = os.path.abspath('')
        ba = sqlite3api.Sqlite3base(None, None, {}, '')
        self.assertEqual(ba._dbdef, {})
        self.assertEqual(ba._sqconn, None)
        self.assertEqual(ba._dbspec, {})
        self.assertEqual(ba._databasefolder, p)
        self.assertEqual(ba._home,
                         os.path.join(p, os.path.basename(p)))
        self.assertIsInstance(ba._control, _sqlite.Sqlite3ControlFile)
        self.assertEqual(len(ba.__dict__), 6)
        self.assertRaisesRegex(
            TypeError,
            "".join((
                "__init__\(\) missing 4 required positional arguments: ",
                "'primary_class', 'secondary_class', ",
                "'database_specification', and 'databasefolder'",
                )),
            sqlite3api.Sqlite3base,
            )
        self.assertRaisesRegex(
            database.DatabaseError,
            "Database folder name None is not valid",
            sqlite3api.Sqlite3base,
            *(None, None, None, None, None))
        self.assertRaisesRegex(
            database.DatabaseError,
            "Database specification must be a dictionary",
            sqlite3api.Sqlite3base,
            *(None, None, None, '', None))
        self.assertRaisesRegex(
            database.DatabaseError,
            "Specification for 't' must be a dictionary",
            sqlite3api.Sqlite3base,
            *(None, None, dict(t=None), '', None))
        self.assertRaisesRegex(
            database.DatabaseError,
            "Specification for t must contain a primary name",
            sqlite3api.Sqlite3base,
            *(None, None, dict(t={}), '', None))
        self.assertRaisesRegex(
            database.DatabaseError,
            "Primary definition key for t must be in fields definition",
            sqlite3api.Sqlite3base,
            *(None,
              None,
              dict(t=dict(primary='key')),
              ''))
        self.assertRaisesRegex(
            database.DatabaseError,
            "Primary definition key for t must be in fields definition",
            sqlite3api.Sqlite3base,
            *(None,
              None,
              dict(t=dict(primary='key', fields=('lock',))),
              ''))
        self.assertRaisesRegex(
            _sqlite.Sqlite3apiError,
            "".join((
                "Cannot construct definition from specification ",
                "because a mandatory argument is missing")),
            sqlite3api.Sqlite3base,
            *(None,
              None,
              filespec.FileSpec(t=dict(primary='key', fields=('lock', 'key'))),
              ''))
        self.assertRaisesRegex(
            sqlite3api.Sqlite3apiError,
            "".join((
                "Cannot construct definition from specification ",
                "because a mandatory argument is missing")),
            sqlite3api.Sqlite3base,
            *(sqlite3api.Sqlite3Primary,
              None,
              filespec.FileSpec(
                  t=dict(
                      primary='key',
                      fields=dict(lock={}, key=True),
                      filedesc=dict(fileorg=None))),
              ''))
        self.assertIsInstance(
            sqlite3api.Sqlite3base(
                sqlite3api.Sqlite3Primary,
                sqlite3api.Sqlite3Secondary,
                filespec.FileSpec(
                    t=dict(
                        primary='key',
                        fields=dict(lock={}, key={}),
                        filedesc=dict(fileorg=None))),
                ''),
            sqlite3api.Sqlite3base)
        self.assertRaisesRegex(
            AttributeError,
            "'NoneType' object has no attribute 'items'",
            sqlite3api.Sqlite3base,
            *(sqlite3api.Sqlite3Primary,
              None,
              dict(
                  t=dict(
                      primary='key',
                      fields=dict(lock={}, key={}),
                      filedesc=dict(fileorg=None),
                      secondary=None)),
              ''))
        self.assertRaisesRegex(
            AttributeError,
            "'int' object has no attribute 'lower'",
            sqlite3api.Sqlite3base,
            *(sqlite3api.Sqlite3Primary,
              None,
              dict(
                  t=dict(
                      primary='key',
                      fields=dict(lock={}, key={}),
                      filedesc=dict(fileorg=None),
                      secondary={1:2})),
              ''))
        self.assertRaisesRegex(
            database.DatabaseError,
            ''.join((
                "Primary definition key for t must not be in ",
                "secondary definition \(ignoring case\)")),
            sqlite3api.Sqlite3base,
            *(sqlite3api.Sqlite3Primary,
              None,
              dict(
                  t=dict(
                      primary='key',
                      fields=dict(lock={}, key={}),
                      filedesc=dict(fileorg=None),
                      secondary=dict(key='key'),
                      )),
              ''))
        self.assertRaisesRegex(
            database.DatabaseError,
            ''.join((
                "Primary definition key for t must not be in ",
                "secondary definition \(ignoring case\)")),
            sqlite3api.Sqlite3base,
            *(sqlite3api.Sqlite3Primary,
              None,
              dict(
                  t=dict(
                      primary='key',
                      fields=dict(lock={}, key={}),
                      filedesc=dict(fileorg=None),
                      secondary=dict(lock='key'),
                      )),
              ''))

        # Now must have FileSpec which brings in some DPT related stuff.
        self.assertRaisesRegex(
            database.DatabaseError,
            ''.join((
                "Primary definition key for t must not be in ",
                "secondary definition \(ignoring case\)")),
            sqlite3api.Sqlite3base,
            *(sqlite3api.Sqlite3Primary,
              None,
              filespec.FileSpec(
                  t=dict(
                      primary='key',
                      fields=dict(lock={}, key={}),
                      filedesc=dict(fileorg=None, brecppg=100),
                      secondary=dict(key=None),
                      btod_factor=5,
                      )),
              ''))
        self.assertRaisesRegex(
            database.DatabaseError,
            ''.join((
                "Primary definition key for t must not be in ",
                "secondary definition \(ignoring case\)")),
            sqlite3api.Sqlite3base,
            *(sqlite3api.Sqlite3Primary,
              None,
              filespec.FileSpec(
                  t=dict(
                      primary='key',
                      fields=dict(lock={}, key={}, Key={}),
                      filedesc=dict(fileorg=None, brecppg=100),
                      secondary=dict(key=None),
                      btod_factor=5,
                      )),
              ''))

        self.assertRaisesRegex(
            database.DatabaseError,
            "".join((
                "Cannot construct definition from specification ",
                "because a mandatory argument is missing")),
            sqlite3api.Sqlite3base,
            *(sqlite3api.Sqlite3Primary,
              None,
              filespec.FileSpec(
                  t=dict(
                      primary='key',
                      fields=dict(lock={}, key={}),
                      filedesc=dict(fileorg=None),
                      secondary=dict(key='lock'),
                      )),
              ''))

    def test___init___02(self):
        api = sqlite3api.Sqlite3base(
            sqlite3api.Sqlite3Primary,
            sqlite3api.Sqlite3Secondary,
            filespec.FileSpec(
                t=dict(
                    primary='key',
                    fields=dict(lock={}, key={}, Skey={}),
                    filedesc=dict(fileorg=None, brecppg=100),
                    secondary=dict(skey=None),
                    btod_factor=5,
                    )),
            '')
        p = os.path.abspath('')
        self.assertEqual(api._databasefolder, p)
        self.assertEqual(api._home, os.path.join(p, os.path.basename(p)))
        self.assertEqual(api._sqconn, None)
        self.assertIsInstance(api._control, _sqlite.Sqlite3ControlFile)
        self.assertEqual(api._dbspec,
                         dict(t={'fields':{'key':{}, 'lock':{}, 'Skey':{}},
                                 'filedesc':{'fileorg':None, 'bsize':2,
                                             'brecppg':100, 'dsize':10},
                                 'primary':'key',
                                 'secondary':{'skey':None},
                                 'btod_factor':5,
                                 'default_records':200,
                                 })
                         )
        self.assertEqual(len(api._dbdef), 1)
        self.assertIsInstance(api._dbdef['t'], _sqlite._Sqlite3Definition)
        df = api._dbdef['t']
        self.assertEqual(df._dbset, 't')
        self.assertEqual(df.dbname_to_secondary_key, {'skey': 'Skey'})
        self.assertIsInstance(df.primary, _sqlite.Sqlite3Primary)
        self.assertIsInstance(df.secondary['Skey'], _sqlite.Sqlite3Secondary)
        self.assertEqual(len(df.secondary), 1)
        self.assertEqual(len(df.__dict__), 4)
        self.assertEqual(len(api.__dict__), 6)


class Sqlite3base_open_context(unittest.TestCase):

    def setUp(self):
        self.api = base('open_context')

    def tearDown(self):
        _delete_folder('open_context')

    def test_open_context_01(self):
        self.assertEqual(self.api._sqconn, None)
        self.assertEqual(os.path.exists(self.api._home), False)
        self.assertEqual(self.api.open_context(), True)
        conn = self.api._sqconn
        self.assertIsInstance(conn, sqlite3.Connection)
        self.assertEqual(os.path.isfile(self.api._home), True)
        self.assertEqual(self.api.open_context(), True)
        self.assertIs(self.api._sqconn, conn)
        self.assertEqual(os.path.isfile(self.api._home), True)


class Sqlite3base_close_context(unittest.TestCase):

    def setUp(self):
        self.api = base('close_context')

    def tearDown(self):
        _delete_folder('close_context')

    def test_close_context_01(self):
        self.assertEqual(self.api._sqconn, None)
        self.assertEqual(self.api.close_context(), None)
        self.assertEqual(self.api._sqconn, None)

    def test_close_context_02(self):
        self.api.open_context()
        conn = self.api._sqconn
        self.assertIsInstance(conn, sqlite3.Connection)
        self.assertEqual(self.api.close_context(), None)
        self.assertIs(self.api._sqconn, conn)
        self.assertEqual(self.api.close_context(), None)
        self.assertIs(self.api._sqconn, conn)


class Sqlite3base_close_database(unittest.TestCase):

    def setUp(self):
        self.api = base('close_database')

    def tearDown(self):
        _delete_folder('close_database')

    def test_close_database_01(self):
        self.assertEqual(self.api._sqconn, None)
        self.assertEqual(self.api.close_database(), None)
        self.assertEqual(self.api._sqconn, None)

    def test_close_database_02(self):
        self.api.open_context()
        self.assertIsInstance(self.api._sqconn, sqlite3.Connection)
        self.assertEqual(self.api.close_database(), None)
        self.assertEqual(self.api._sqconn, None)
        self.assertEqual(self.api.close_database(), None)
        self.assertEqual(self.api._sqconn, None)


class Sqlite3base_backout(unittest.TestCase):
    # Nothing to backout, just check the calls work.

    def setUp(self):
        self.api = base('backout')

    def tearDown(self):
        _delete_folder('backout')

    def test_backout_01(self):
        self.assertEqual(self.api._sqconn, None)
        self.assertEqual(self.api.backout(), None)
        self.api.open_context()
        self.assertIsInstance(self.api._sqconn, sqlite3.Connection)
        self.api.start_transaction()
        self.assertEqual(self.api.backout(), None)
        self.assertIsInstance(self.api._sqconn, sqlite3.Connection)


class Sqlite3base_close_contexts(unittest.TestCase):

    def setUp(self):
        self.api = base('close_contexts')

    def tearDown(self):
        _delete_folder('close_contexts')

    def test_close_contexts_01(self):
        self.assertEqual(self.api.close_contexts('DPT compatibility'), None)


class Sqlite3base_commit(unittest.TestCase):
    # Nothing to commit, just check the calls work.

    def setUp(self):
        self.api = base('commit')

    def tearDown(self):
        _delete_folder('commit')

    def test_commit_01(self):
        self.assertEqual(self.api._sqconn, None)
        self.assertEqual(self.api.commit(), None)
        self.api.open_context()
        self.assertIsInstance(self.api._sqconn, sqlite3.Connection)
        self.api.start_transaction()
        self.assertEqual(self.api.commit(), None)
        self.assertIsInstance(self.api._sqconn, sqlite3.Connection)


class Sqlite3base_db_compatibility_hack(unittest.TestCase):

    def setUp(self):
        self.api = base('db_compatibility_hack')

    def tearDown(self):
        _delete_folder('db_compatibility_hack')

    def test_db_compatibility_hack_01(self):
        record = ('key', 'value')
        srkey = self.api.encode_record_number(3456)
        self.assertEqual(
            self.api.db_compatibility_hack(record, srkey), record)
        record = ('key', None)
        value = self.api.decode_record_number(srkey)
        self.assertEqual(
            self.api.db_compatibility_hack(record, srkey),
            ('key', value))


class Sqlite3base_exists(unittest.TestCase):

    def setUp(self):
        self.api = base('exists')

    def tearDown(self):
        _delete_folder('exists')

    def test_exists_01(self):
        self.assertRaisesRegex(
            KeyError,
            "None",
            self.api.exists,
            *(None, None))
        self.assertEqual(self.api.exists('games', None), False)
        self.assertEqual(self.api.exists('games', 'Site'), True)


class Sqlite3base_files_exist(unittest.TestCase):

    def setUp(self):
        self.api = base('files_exist')

    def tearDown(self):
        _delete_folder('files_exist')

    def test_exists_01(self):
        self.assertEqual(self.api.files_exist(), False)
        self.api.open_context()
        self.assertEqual(self.api.files_exist(), True)
        self.api.close_database()
        self.assertEqual(self.api.files_exist(), True)


class Sqlite3base_make_cursor(unittest.TestCase):

    def setUp(self):
        self.api = base('database_cursor')

    def tearDown(self):
        _delete_folder('database_cursor')

    def test_make_cursor_01(self):
        self.assertRaisesRegex(
            KeyError,
            "None",
            self.api.database_cursor,
            *(None, None))
        self.assertIsInstance(
            self.api.database_cursor('games', 'Site'),
            _sqlite.CursorSqlite3Secondary)
        self.assertIsInstance(
            self.api.database_cursor('games', 'games'),
            _sqlite.CursorSqlite3Primary)


class Sqlite3base_get_database_folder(unittest.TestCase):

    def setUp(self):
        self.api = base('get_database_folder')

    def tearDown(self):
        _delete_folder('get_database_folder')

    def test_get_database_folder_01(self):
        self.assertEqual(
            self.api.get_database_folder(),
            os.path.expanduser(
                os.path.join('~', 'sqlite3api_tests', 'get_database_folder')))


class Sqlite3base_get_database(unittest.TestCase):

    def setUp(self):
        self.api = base('get_database')

    def tearDown(self):
        _delete_folder('get_database')

    def test_get_database_01(self):
        self.assertRaisesRegex(
            KeyError,
            "None",
            self.api.get_database,
            *(None, None))
        self.assertEqual(self.api.get_database('games', 'Site'), None)
        self.assertEqual(self.api.get_database('games', 'games'), None)
        self.api.open_context()
        self.assertIsInstance(
            self.api.get_database('games', 'Site'),
            sqlite3.Connection)
        self.assertIsInstance(
            self.api.get_database('games', 'games'),
            sqlite3.Connection)


class Sqlite3base_get_database_instance(unittest.TestCase):

    def setUp(self):
        self.api = base('get_database_instance')

    def tearDown(self):
        _delete_folder('get_database_instance')

    def test_get_database_instance_01(self):
        self.assertRaisesRegex(
            KeyError,
            "None",
            self.api.get_database_instance,
            *(None, None))
        self.assertIsInstance(
            self.api.get_database_instance('games', 'Site'),
            sqlite3api.Sqlite3Secondary)
        self.assertIsInstance(
            self.api.get_database_instance('games', 'games'),
            sqlite3api.Sqlite3Primary)


class Sqlite3base_get_first_primary_key_for_index_key(unittest.TestCase):
    # Test results are suspicious but may be correct.

    def setUp(self):
        self.api = base('get_first_primary_key_for_index_key')

    def tearDown(self):
        _delete_folder('get_first_primary_key_for_index_key')

    def test_get_first_primary_key_for_index_key_01(self):
        self.assertRaisesRegex(
            KeyError,
            "None",
            self.api.get_first_primary_key_for_index_key,
            *(None, None, None))
        self.assertRaisesRegex(
            TypeError,
            "'NoneType' object is not subscriptable",
            self.api.get_first_primary_key_for_index_key,
            *('games', 'Site', 'k'))
        self.assertRaisesRegex(
            sqlite3api.Sqlite3apiError,
            "get_first_primary_key_for_index_key for primary table",
            self.api.get_first_primary_key_for_index_key,
            *('games', 'games', 'k'))
        self.api.open_context()
        #self.assertEqual(
        #    self.api.get_first_primary_key_for_index_key('games', 'games', 'k'),
        #    None)
        self.assertRaisesRegex(
            sqlite3api.Sqlite3apiError,
            "get_first_primary_key_for_index_key for primary table",
            self.api.get_first_primary_key_for_index_key,
            *('games', 'games', 'k'))


class Sqlite3base_get_primary_record(unittest.TestCase):

    def setUp(self):
        self.api = base('get_primary_record')

    def tearDown(self):
        _delete_folder('get_primary_record')

    def test_get_primary_record_01(self):
        self.assertRaisesRegex(
            KeyError,
            "None",
            self.api.get_primary_record,
            *(None, None))
        self.assertRaisesRegex(
            TypeError,
            "'NoneType' object is not subscriptable",
            self.api.get_primary_record,
            *('games', 'k'))
        self.api.open_context()
        self.assertEqual(self.api.get_primary_record('games', 'k'), None)


class Sqlite3base_is_primary(unittest.TestCase):

    def setUp(self):
        self.api = base('is_primary')

    def tearDown(self):
        _delete_folder('is_primary')

    def test_is_primary_01(self):
        self.assertRaisesRegex(
            KeyError,
            "None",
            self.api.is_primary,
            *(None, None))
        self.assertEqual(self.api.is_primary('games', 'games'), True)
        self.assertEqual(self.api.is_primary('games', 'Site'), False)


class Sqlite3base_is_primary_recno(unittest.TestCase):

    def setUp(self):
        self.api = base('is_primary_recno')

    def tearDown(self):
        _delete_folder('is_primary_recno')

    def test_is_primary_recno_01(self):
        self.assertEqual(
            self.api.is_primary_recno('Berkeley DB compatibility'), True)


class Sqlite3base_is_recno(unittest.TestCase):

    def setUp(self):
        self.api = base('is_recno')

    def tearDown(self):
        _delete_folder('is_recno')

    def test_is_recno_01(self):
        self.assertRaisesRegex(
            KeyError,
            "None",
            self.api.is_recno,
            *(None, None))
        self.assertEqual(self.api.is_recno('games', 'games'), True)
        self.assertEqual(self.api.is_recno('games', 'Site'), False)


class Sqlite3base_open_contexts(unittest.TestCase):

    def setUp(self):
        self.api = base('open_contexts')

    def tearDown(self):
        _delete_folder('open_contexts')

    def test_open_contexts_01(self):
        self.assertEqual(self.api.open_contexts('DPT compatibility'), None)


class Sqlite3base_get_packed_key(unittest.TestCase):

    def setUp(self):
        self.api = base('get_packed_key')

    def tearDown(self):
        _delete_folder('get_packed_key')

    def test_get_packed_key_01(self):
        instance = record.Record(record.Key, record.Value)
        self.assertEqual(
            self.api.get_packed_key('Berkeley DB compatibility', instance),
            instance.key.pack())


class Sqlite3base_decode_as_primary_key(unittest.TestCase):

    def setUp(self):
        self.api = base('decode_as_primary_key')

    def tearDown(self):
        _delete_folder('decode_as_primary_key')

    def test_decode_as_primary_key_01(self):
        self.assertEqual(self.api.decode_as_primary_key(
            'Ignored in sqlite3', 10), 10)
        srkey = self.api.encode_record_number(3456)
        self.assertEqual(self.api.decode_as_primary_key(
            'Ignored in sqlite3', srkey), 3456)


class Sqlite3base_encode_primary_key(unittest.TestCase):

    def setUp(self):
        self.api = base('encode_primary_key')

    def tearDown(self):
        _delete_folder('encode_primary_key')

    def test_encode_primary_key_01(self):
        instance = record.Record(record.KeyData, record.Value)
        instance.key.load(23)
        self.assertEqual(
            self.api.encode_primary_key('games', instance),
            '23')


class Sqlite3base_use_deferred_update_process(unittest.TestCase):

    def setUp(self):
        self.api = base('use_deferred_update_process')

    def tearDown(self):
        _delete_folder('use_deferred_update_process')

    def test_use_deferred_update_process_01(self):
        self.assertRaisesRegex(
            database.DatabaseError,
            "use_deferred_update_process not implemented",
            self.api.use_deferred_update_process,
            **{'zzz': 2, 'd': None})


class Sqlite3base_initial_database_size(unittest.TestCase):
    # initial_database_size does nothing, it is for DPT compatibility.

    def setUp(self):
        self.api = base('initial_database_size')

    def tearDown(self):
        _delete_folder('initial_database_size')

    def test_initial_database_size_01(self):
        self.assertEqual(self.api.initial_database_size(), True)


class Sqlite3base_increase_database_size(unittest.TestCase):

    def setUp(self):
        self.api = base('increase_database_size')

    def tearDown(self):
        _delete_folder('increase_database_size')

    def test_increase_database_size_01(self):
        self.assertEqual(
            self.api.increase_database_size(**{'l':'DPT compatibility'}), None)


class Sqlite3base_do_database_task(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        _delete_folder('do_database_task')

    def test_do_database_task_01(self):
        def f(a, b, **c):
            pass
        self.api = base('do_database_task')
        self.assertRaisesRegex(
            TypeError,
            ''.join((
                "__init__\(\) missing 3 required positional arguments: ",
                "'secondary_class', 'database_specification', ",
                "and 'databasefolder'",
                )),
            self.api.do_database_task,
            *(f,))

    def test_do_database_task_02(self):
        def f(a, b, **c):
            pass
        class C(sqlite3api.Sqlite3base):
            def __init__(self, *a, **k):
                super().__init__(
                    sqlite3api.Sqlite3Primary,
                    sqlite3api.Sqlite3Secondary,
                    samplefilespec(),
                    *a,
                    **k)
        self.api = C(
            os.path.expanduser(
                os.path.join('~', 'sqlite3api_tests', 'do_database_task')))
        self.assertEqual(self.api.do_database_task(f), None)


class Sqlite3base_make_recordset_key(unittest.TestCase):

    def setUp(self):
        self.api = base('make_recordset_key')

    def tearDown(self):
        _delete_folder('make_recordset_key')

    def test_make_recordset_key_01(self):
        self.assertRaisesRegex(
            KeyError,
            "None",
            self.api.make_recordset_key,
            *(None, None))
        self.assertRaisesRegex(
            TypeError,
            "'NoneType' object is not subscriptable",
            self.api.make_recordset_key,
            *('games', 'games'))
        self.api.open_context()
        self.assertRaisesRegex(
            TypeError,
            "object of type 'NoneType' has no len\(\)",
            self.api.make_recordset_key,
            *('games', 'games'))
        self.assertIsInstance(
            self.api.make_recordset_key('games', 'Site'),
            recordset.Recordset)


class Sqlite3base_make_recordset_key_startswith(unittest.TestCase):

    def setUp(self):
        self.api = base('make_recordset_key_startswith')

    def tearDown(self):
        _delete_folder('make_recordset_key_startswith')

    def test_make_recordset_key_startswith_01(self):
        self.assertRaisesRegex(
            KeyError,
            "None",
            self.api.make_recordset_key_startswith,
            *(None, None))
        self.assertRaisesRegex(
            sqlite3api.Sqlite3apiError,
            ''.join((
                "populate_recordset_key_startswith not available on ",
                "primary database")),
            self.api.make_recordset_key_startswith,
            *('games', 'games'))
        if sys.version_info[:2] < (3,4):
            self.assertRaisesRegex(
                TypeError,
                'sequence item 0: expected bytes, NoneType found',
                self.api.make_recordset_key_startswith,
                *('games', 'Site'))
        else:
            self.assertRaisesRegex(
                TypeError,
                "'NoneType' object is not subscriptable",
                self.api.make_recordset_key_startswith,
                *('games', 'Site'))
        self.api.open_context()
        self.assertRaisesRegex(
            sqlite3api.Sqlite3apiError,
            ''.join((
                "populate_recordset_key_startswith not available on ",
                "primary database")),
            self.api.make_recordset_key_startswith,
            *('games', 'games'))
        if sys.version_info[:2] < (3,4):
            self.assertRaisesRegex(
                TypeError,
                'sequence item 0: expected bytes, NoneType found',
                self.api.make_recordset_key_startswith,
                *('games', 'Site'))
        else:
            self.assertRaisesRegex(
                TypeError,
                'sequence item 0: expected a bytes-like object, NoneType found',
                self.api.make_recordset_key_startswith,
                *('games', 'Site'))


class Sqlite3base_make_recordset_key_range(unittest.TestCase):

    def setUp(self):
        self.api = base('make_recordset_key_range')

    def tearDown(self):
        _delete_folder('make_recordset_key_range')

    def test_make_recordset_key_range_01(self):
        self.assertRaisesRegex(
            KeyError,
            "None",
            self.api.make_recordset_key_range,
            *(None, None))
        self.assertRaisesRegex(
            AttributeError,
            "'NoneType' object has no attribute 'cursor'",
            self.api.make_recordset_key_range,
            *('games', 'games'))
        self.assertRaisesRegex(
            TypeError,
            "'NoneType' object is not subscriptable",
            self.api.make_recordset_key_range,
            *('games', 'Site'))
        self.api.open_context()
        self.assertIsInstance(
            self.api.make_recordset_key_range('games', 'games'),
            recordset.Recordset)
        self.assertIsInstance(
            self.api.make_recordset_key_range('games', 'Site'),
            recordset.Recordset)


class Sqlite3base_make_recordset_all(unittest.TestCase):

    def setUp(self):
        self.api = base('make_recordset_all')

    def tearDown(self):
        _delete_folder('make_recordset_all')

    def test_make_recordset_all_01(self):
        self.assertRaisesRegex(
            KeyError,
            "None",
            self.api.make_recordset_all,
            *(None, None))
        self.assertRaisesRegex(
            AttributeError,
            "'NoneType' object has no attribute 'cursor'",
            self.api.make_recordset_all,
            *('games', 'games'))
        self.assertRaisesRegex(
            TypeError,
            "'NoneType' object is not subscriptable",
            self.api.make_recordset_all,
            *('games', 'Site'))
        self.api.open_context()
        self.assertIsInstance(
            self.api.make_recordset_all('games', 'games'),
            recordset.Recordset)
        self.assertIsInstance(
            self.api.make_recordset_all('games', 'Site'),
            recordset.Recordset)


class Sqlite3base_file_records_under(unittest.TestCase):

    def setUp(self):
        self.api = base('file_records_under')

    def tearDown(self):
        _delete_folder('file_records_under')

    def test_file_records_under_01(self):
        self.assertRaisesRegex(
            AttributeError,
            "'NoneType' object has no attribute 'dbidentity'",
            self.api.file_records_under,
            *(None, None, None, None))
        rs = recordset.Recordset(dbhome=self.api, dbset='games')
        self.assertRaisesRegex(
            KeyError,
            "None",
            self.api.file_records_under,
            *(None, None, rs, None))
        self.assertRaisesRegex(
            database.DatabaseError,
            "file_records_under not implemented for Sqlite3Primary",
            self.api.file_records_under,
            *('games', 'games', rs, b'dd'))
        self.assertRaisesRegex(
            TypeError,
            "'NoneType' object is not subscriptable",
            self.api.file_records_under,
            *('games', 'Site', rs, b'dd'))
        self.api.open_context()
        self.assertRaisesRegex(
            database.DatabaseError,
            "Record set was not created from this database instance",
            self.api.file_records_under,
            *('games', 'games', rs, b'dd'))
        rs = recordset.Recordset(dbhome=self.api, dbset='games')
        self.assertRaisesRegex(
            database.DatabaseError,
            "file_records_under not implemented for Sqlite3Primary",
            self.api.file_records_under,
            *('games', 'games', rs, b'dd'))
        rs._dbset = 'sssss'
        self.assertRaisesRegex(
            database.DatabaseError,
            "Record set was not created from dbset database",
            self.api.file_records_under,
            *('games', 'games', rs, b'dd'))
        rs = recordset.Recordset(dbhome=self.api, dbset='games')
        self.assertEqual(
            self.api.file_records_under('games', 'Site', rs, b'dd'),
            None)


class Sqlite3base_method_classes(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_method_classes_01(self):
        bit = sqlite3api.Sqlite3api
        api = sqlite3api.Sqlite3base
        base = database.Database
        enc = _sqlite._DatabaseEncoders

        # Differences from dbapi_test are noted by '#'.
        self.assertEqual(
            [a for a in dir(bit)
                 if not a.startswith('__') and not a.endswith('__')],
            ['allocate_and_open_contexts',
             'backout',
             'cede_contexts_to_process',
             'close_context',
             'close_contexts',
             'close_database',
             'commit',
             'connection', #add
             'create_recordset_cursor',
             'database_cursor',
             'database_specification',
             'db_compatibility_hack',
             'decode_as_primary_key',
             'decode_record_number',
             'deferred_update_housekeeping',
             'delete_instance',
             'do_database_task',
             'do_deferred_updates',
             'edit_instance',
             'encode_primary_key',
             'encode_record_key',
             'encode_record_number',
             'encode_record_selector',
             'exists',
             'file_records_under',
             'files_exist',
             'get_associated_indicies',
             'get_database',
             'get_database_filename', # trailing 's' removed
             'get_database_folder',
             'get_database_instance',
             'get_dbsets',
             'get_first_primary_key_for_index_key',
             'get_packed_key',
             'get_primary_record',
             'get_table_index',
             'get_table_indicies',
             #'get_transaction',
             'increase_database_size',
             'initial_database_size',
             'is_primary',
             'is_primary_recno',
             'is_recno',
             'make_connection', # add
             'make_recordset',
             'make_recordset_all',
             'make_recordset_key',
             'make_recordset_key_like',
             'make_recordset_key_range',
             'make_recordset_key_startswith',
             'make_root',
             'open_context',
             'open_contexts',
             'put_instance',
             'record_finder',
             'record_selector',
             'repair_cursor',
             'set_defer_update',
             'start_transaction',
             'unfile_records_under',
             'unset_defer_update',
             'use_deferred_update_process',
             'values_finder',
             'values_selector'])
        self.assertEqual(
            len([a for a in dir(bit)
                 if not a.startswith('__') and not a.endswith('__')]),
            63)
        self.assertEqual(
            len([a for a in dir(api)
                 if not a.startswith('__') and not a.endswith('__')]),
            63)
        self.assertEqual(
            len([a for a in dir(base)
                 if not a.startswith('__') and not a.endswith('__')]),
            49)
        self.assertEqual(
            len([a for a in dir(enc)
                 if not a.startswith('__') and not a.endswith('__')]),
            4)

        # Methods inherited from Sqlite3base
        self.assertIs(api.backout, bit.backout)
        self.assertIs(api.close_contexts, bit.close_contexts)
        self.assertIs(api.close_database, bit.close_database)
        self.assertIs(api.commit, bit.commit)
        self.assertIs(api.db_compatibility_hack, bit.db_compatibility_hack)
        self.assertIs(api.exists, bit.exists)
        self.assertIs(api.files_exist, bit.files_exist)
        self.assertIs(api.database_cursor, bit.database_cursor)
        self.assertIs(api.repair_cursor, bit.repair_cursor)
        self.assertIs(api.get_database_folder, bit.get_database_folder)
        self.assertIs(api.get_database, bit.get_database)
        self.assertIs(api.get_database_instance, bit.get_database_instance)
        self.assertIs(
            api.get_first_primary_key_for_index_key,
            bit.get_first_primary_key_for_index_key)
        self.assertIs(api.get_primary_record, bit.get_primary_record)
        self.assertIs(api.is_primary, bit.is_primary)
        self.assertIs(api.is_primary_recno, bit.is_primary_recno)
        self.assertIs(api.is_recno, bit.is_recno)
        self.assertIs(api.open_contexts, bit.open_contexts)
        self.assertIs(api.allocate_and_open_contexts,
                      bit.allocate_and_open_contexts)
        self.assertIs(api.get_packed_key, bit.get_packed_key)
        self.assertIs(api.decode_as_primary_key, bit.decode_as_primary_key)
        self.assertIs(api.encode_primary_key, bit.encode_primary_key)
        self.assertIs(
            api.use_deferred_update_process,
            bit.use_deferred_update_process)
        self.assertIs(api.initial_database_size, bit.initial_database_size)
        self.assertIs(api.increase_database_size, bit.increase_database_size)
        self.assertIs(api.do_database_task, bit.do_database_task)
        self.assertIs(api.make_recordset_key, bit.make_recordset_key)
        self.assertIs(
            api.make_recordset_key_startswith,
            bit.make_recordset_key_startswith)
        self.assertIs(
            api.make_recordset_key_range,
            bit.make_recordset_key_range)
        self.assertIs(api.make_recordset_all, bit.make_recordset_all)
        self.assertIs(api.file_records_under, bit.file_records_under)
        #self.assertIs(api.is_engine_uses_bytes, bit.is_engine_uses_bytes)
        #self.assertIs(api.is_engine_uses_str, bit.is_engine_uses_str)
        self.assertIs(api.start_transaction, bit.start_transaction)
        self.assertIs(api.cede_contexts_to_process,
                      bit.cede_contexts_to_process)
        
        # Methods inherited from _DatabaseEncoders
        self.assertIs(enc.encode_record_number, bit.encode_record_number)
        self.assertIs(enc.decode_record_number, bit.decode_record_number)
        self.assertIs(enc.encode_record_selector, bit.encode_record_selector)
        self.assertIs(enc.encode_record_key, bit.encode_record_key)
        
        # Methods inherited from Database
        self.assertIs(base.deferred_update_housekeeping,
                      bit.deferred_update_housekeeping)

        # Sqlite3base methods overridden
        self.assertIsNot(api.__init__, bit.__init__)
        self.assertIsNot(api.open_context, bit.open_context)
        self.assertIsNot(api.close_context, bit.close_context)
        self.assertIs(api.delete_instance, bit.delete_instance)
        self.assertIs(api.edit_instance, bit.edit_instance)
        self.assertIs(api.put_instance, bit.put_instance)
        self.assertIsNot(api.do_deferred_updates, bit.do_deferred_updates)
        self.assertIs(api.set_defer_update, bit.set_defer_update)
        self.assertIs(api.unset_defer_update, bit.unset_defer_update)


class Sqlite3api_init(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test___init___01(self):
        self.assertRaisesRegex(
            TypeError,
            "".join((
                "__init__\(\) missing 1 required positional argument: ",
                "'database_specification'",
                )),
            sqlite3api.Sqlite3api,
            )
        self.assertRaisesRegex(
            database.DatabaseError,
            "Database folder name None is not valid",
            sqlite3api.Sqlite3api,
            *(samplefilespec(),),
            **dict(databasefolder=None)
            )
        self.assertRaisesRegex(
            database.DatabaseError,
            "Database specification must be a dictionary",
            sqlite3api.Sqlite3api,
            *(None,),
            **dict(databasefolder='')
            )

    def test___init___02(self):
        api = sqlite3api.Sqlite3api(
            filespec.FileSpec(
                t=dict(
                    primary='key',
                    fields=dict(lock={}, key={}, Skey={}),
                    filedesc=dict(fileorg=None),
                    secondary=dict(skey=None),
                    )),
            databasefolder='')
        p = os.path.abspath('')
        self.assertEqual(api._databasefolder, p)
        self.assertEqual(api._home, os.path.join(p, os.path.basename(p)))
        self.assertEqual(api._sqconn, None)
        self.assertIsInstance(api._control, _sqlite.Sqlite3ControlFile)
        self.assertEqual(api._dbspec,
                         dict(t={'fields':{'key':{}, 'lock':{}, 'Skey':{}},
                                 'filedesc':{'fileorg':None, 'bsize':20,
                                             'brecppg':10, 'dsize':160},
                                 'primary':'key',
                                 'secondary':{'skey':None},
                                 'btod_factor':8,
                                 'default_records':200,
                                 })
                         )
        self.assertEqual(len(api._dbdef), 1)
        self.assertIsInstance(api._dbdef['t'], _sqlite._Sqlite3Definition)
        df = api._dbdef['t']
        self.assertEqual(df._dbset, 't')
        self.assertEqual(df.dbname_to_secondary_key, {'skey': 'Skey'})
        self.assertIsInstance(df.primary, _sqlite.Sqlite3Primary)
        self.assertIsInstance(df.secondary['Skey'], _sqlite.Sqlite3Secondary)
        self.assertEqual(len(df.secondary), 1)
        self.assertEqual(len(df.__dict__), 4)
        self.assertEqual(len(api.__dict__), 6)


class Sqlite3base_set_defer_update(unittest.TestCase):

    def setUp(self):
        self.api = base('set_defer_update')

    def tearDown(self):
        _delete_folder('set_defer_update')

    def test_set_defer_update_01(self):
        self.assertEqual(self.api.set_defer_update(), False)
        self.assertEqual(self.api.set_defer_update(duallowed=True), True)
        self.assertEqual(self.api.set_defer_update(duallowed='xx'), 'xx')
        self.assertEqual(
            self.api.set_defer_update(db='Berkeley DB compatibility'), False)


class Sqlite3base_unset_defer_update(unittest.TestCase):

    def setUp(self):
        self.api = base('unset_defer_update')

    def tearDown(self):
        _delete_folder('unset_defer_update')

    def test_unset_defer_update_01(self):
        self.assertEqual(
            self.api.unset_defer_update(db='Berkeley DB compatibility'), True)


class Sqlite3base_do_deferred_updates(unittest.TestCase):

    def setUp(self):
        self.api = base('do_deferred_updates')

    def tearDown(self):
        _delete_folder('do_deferred_updates')

    def test_do_deferred_updates_01(self):
        self.api.open_context()
        self.assertRaisesRegex(
            sqlite3api.Sqlite3apiError,
            "'pyscript' is not an existing file",
            self.api.do_deferred_updates,
            *('pyscript', 'filepath'))
        self.assertRaisesRegex(
            sqlite3api.Sqlite3apiError,
            "'filepath' is not an existing file",
            self.api.do_deferred_updates,
            *(self.api._home, 'filepath'))
        self.assertRaisesRegex(
            sqlite3api.Sqlite3apiError,
            "'filepath' is not an existing file",
            self.api.do_deferred_updates,
            *(self.api._home, (self.api._home, 'filepath')))
        script = os.path.join(os.path.dirname(self.api._home), 'script')
        open(script, 'w').close()
        sp = self.api.do_deferred_updates(script, self.api._home)
        self.assertIsInstance(sp, subprocess.Popen)
        sp.wait()


class Sqlite3api_put_instance(unittest.TestCase):

    def setUp(self):
        self.api = api('put_instance')
        self.sr = SampleRecord()

    def tearDown(self):
        self.api.close_database()
        _delete_folder('put_instance')

    def test_put_instance_01(self):
        self.assertRaisesRegex(
            AttributeError,
            "'NoneType' object has no attribute 'key'",
            self.api.put_instance,
            *(None, None))
        instance = record.Record(record.KeyData, record.Value)
        instance.key.recno = 0
        self.assertRaisesRegex(
            KeyError,
            "None",
            self.api.put_instance,
            *(None, instance))
        instance = record.Record(record.KeyData, record.Value)
        instance.key.recno = 0
        self.assertRaisesRegex(
            AttributeError,
            "'NoneType' object has no attribute 'cursor'",
            self.api.put_instance,
            *('games', instance))

    def test_put_instance_02(self):
        instance = record.Record(record.KeyData, record.Value)
        instance.key.recno = 0
        self.api.open_context()
        self.assertEqual(self.api.put_instance('games', instance), None)

    def test_put_instance_03(self):
        instance = record.Record(record.KeyData, record.Value)
        instance.key.recno = 4
        self.api.open_context()
        self.assertEqual(self.api.put_instance('games', instance), None)
        self.assertEqual(self.api.put_instance('games', instance), None)

    def test_put_instance_04(self):
        instance = record.Record(record.KeyData, _Value)
        instance.key.recno = 0
        self.api.open_context()
        self.assertRaisesRegex(
            KeyError,
            "'hhhhhh'",
            self.api.put_instance,
            *('games', instance))

    def test_put_instance_05(self):
        database_interface.test_put_instance_05(self,
                                                collect_counts,
                                                game_number_to_record_number)

    def test_put_instance_06(self):
        database_interface.test_put_instance_06(self,
                                                collect_counts,
                                                game_number_to_record_number)

    def test_put_instance_07(self):
        database_interface.test_put_instance_07(self,
                                                collect_counts,
                                                game_number_to_record_number)

    def test_put_instance_08(self):
        database_interface.test_put_instance_08(self,
                                                collect_counts,
                                                game_number_to_record_number)

    def test_put_instance_09(self):
        database_interface.test_put_instance_09(self,
                                                collect_counts,
                                                game_number_to_record_number)

    def test_put_instance_10(self):
        database_interface.test_put_instance_10(self,
                                                collect_counts,
                                                game_number_to_record_number)


class Sqlite3api_delete_instance(unittest.TestCase):

    def setUp(self):
        self.api = api('delete_instance')
        self.sr = SampleRecord()

    def tearDown(self):
        self.api.close_database()
        _delete_folder('delete_instance')

    def test_delete_instance_01(self):
        self.assertRaisesRegex(
            AttributeError,
            "'NoneType' object has no attribute 'key'",
            self.api.delete_instance,
            *(None, None))
        instance = record.Record(record.KeyData, record.Value)
        instance.key.recno = 2
        self.assertRaisesRegex(
            KeyError,
            "None",
            self.api.delete_instance,
            *(None, instance))
        instance = record.Record(record.KeyData, record.Value)
        instance.key.recno = 2
        self.assertRaisesRegex(
            TypeError,
            "'NoneType' object is not subscriptable",
            self.api.delete_instance,
            *('games', instance))

    def test_delete_instance_02(self):
        instance = record.Record(record.KeyData, record.Value)
        instance.key.recno = 2
        self.api.open_context()
        self.assertRaisesRegex(
            sqlite3api.Sqlite3apiError,
            "Existence bit map for segment does not exist",
            self.api.delete_instance,
            *('games', instance))

    def test_delete_instance_03(self):
        instance = record.Record(record.KeyData, record.Value)
        instance.key.recno = 0
        self.api.open_context()
        self.api.put_instance('games', instance)
        instance = record.Record(record.KeyData, _Value)
        instance.key.recno = 100000
        self.assertRaisesRegex(
            sqlite3api.Sqlite3apiError,
            "Existence bit map for segment does not exist",
            self.api.delete_instance,
            *('games', instance))
        instance = record.Record(record.KeyData, record.Value)
        instance.key.recno = 10000
        self.assertEqual(self.api.delete_instance('games', instance), None)
        instance = record.Record(record.KeyData, record.Value)
        instance.key.recno = 1
        self.assertEqual(self.api.delete_instance('games', instance), None)
        instance = record.Record(record.KeyData, record.Value)
        instance.key.recno = 1
        self.assertEqual(self.api.delete_instance('games', instance), None)

    def test_delete_instance_04(self):
        database_interface.test_delete_instance_01(self,
                                                   collect_counts,
                                                   game_number_to_record_number)

    def test_delete_instance_05(self):
        database_interface.test_delete_instance_02(self,
                                                   collect_counts,
                                                   game_number_to_record_number)


class Sqlite3api_edit_instance(unittest.TestCase):

    def setUp(self):
        self.api = api('edit_instance')
        self.sr = SampleRecord()

    def tearDown(self):
        self.api.close_database()
        _delete_folder('edit_instance')

    def test_edit_instance_01(self):
        self.assertRaisesRegex(
            AttributeError,
            "'NoneType' object has no attribute 'key'",
            self.api.edit_instance,
            *(None, None))
        instance = record.Record(record.KeyData, record.Value)
        instance.key.recno = 2
        edited = record.Record(record.KeyData, record.Value)
        edited.key.recno = 2
        instance.newrecord = edited
        self.assertRaisesRegex(
            KeyError,
            "None",
            self.api.edit_instance,
            *(None, instance))

    def test_edit_instance_02(self):
        instance = record.Record(record.KeyData, record.Value)
        instance.key.recno = 2
        edited = record.Record(record.KeyData, record.Value)
        edited.key.recno = 2
        instance.newrecord = edited
        self.assertEqual(self.api.edit_instance('games', instance), None)

    def test_edit_instance_03(self):
        instance = record.Record(record.KeyData, record.Value)
        instance.key.recno = 2
        edited = record.Record(record.KeyData, _ValueEdited)
        edited.key.recno = 2
        instance.newrecord = edited
        self.api.open_context()
        self.assertRaisesRegex(
            KeyError,
            "'hhhhhh'",
            self.api.edit_instance,
            *('games', instance))

    def test_edit_instance_04(self):
        instance = record.Record(record.KeyData, record.Value)
        instance.key.recno = 2
        edited = record.Record(record.KeyData, _ValueEdited)
        edited.key.recno = 2
        self.api.open_context()
        self.api.put_instance('games', instance)
        instance.newrecord = edited
        self.assertRaisesRegex(
            KeyError,
            "'hhhhhh'",
            self.api.edit_instance,
            *('games', instance))

    def test_edit_instance_05(self):
        instance = record.Record(record.KeyData, record.Value)
        instance.key.recno = 2
        edited = record.Record(record.KeyData, _ValueEdited)
        edited.key.recno = 3
        self.api.open_context()
        self.api.put_instance('games', instance)
        instance.newrecord = edited
        self.assertRaisesRegex(
            KeyError,
            "'hhhhhh'",
            self.api.edit_instance,
            *('games', instance))

    def test_edit_instance_06(self):
        database_interface.test_edit_instance_06(self,
                                                 collect_counts,
                                                 game_number_to_record_number)

    def test_edit_instance_07(self):
        database_interface.test_edit_instance_07(self,
                                                 collect_counts,
                                                 game_number_to_record_number)


class Sqlite3api_put_instance_with_holes(unittest.TestCase):

    def setUp(self):

        # DPT segment size is 65280 (8k bytes - 32 reserved) while apsw, bsddb3,
        # and sqlite3, segment size is 32768 (4k bytes).  Reuse record number
        # tests cannot produce the same answer in general on the two segment
        # sizes, even if the DPT algorithm deciding when to fill a hole is the
        # one used for apsw, bsddb3, and sqlite3.
        # The api() function calls samplefilespec with the default fileorg
        # argument because DPT is not the database engine.
        self.dpt_reuse_record_numbers = False
        
        self.api = api('put_instance_holes')
        self.sr = SampleRecord()

    def tearDown(self):
        self.api.close_database()
        _delete_folder('put_instance_holes')

    def test_holes_01(self):
        database_interface.test_holes_01(self,
                                         collect_counts,
                                         game_number_to_record_number)


def game_number_to_record_number(game_number):
    # Record numbers start at 1 and game numbers start at 1
    return game_number


def record_number_to_game_number(record_number):
    # Record numbers start at 1 and game numbers start at 1
    return record_number


def collect_counts(testcase,
                   record_adder,
                   data_adder,
                   answers,
                   counts,
                   segments,
                   recordsets,
                   rowexceptions):
    db_directory = testcase.api.get_database_folder()
    conn = sqlite3.connect(
        os.path.join(db_directory, os.path.split(db_directory)[-1]))
    cursor = conn.cursor()
    filespec = samplefilespec()
    primaries = {filespec[t][PRIMARY]:t for t in filespec}
    for t in answers['data']:
        testcase.assertEqual(t in primaries, True)
        segments[t] = dict()
        rowexceptions[t] = dict()
        cursor.execute(t.join(('select * from ', '')))
        while True:
            r = cursor.fetchone()
            if not r:
                break
            # Allow for place holder null record for reuse record number
            if r[1] is None:
                continue
            data_adder(rowexceptions[t],
                       segments[t],
                       (r[0], r[1]),
                       answers['data'][t]['defaultrow'],
                       record_number_to_game_number)
    for t in answers['data']:
        for u in answers['records']:
            if u not in filespec[primaries[t]][SECONDARY]:
                continue
            segments[u] = dict()
            recordsets[u] = dict()
            cursor.execute(u.join(('select * from ', '')))
            segment_cursor = conn.cursor()
            while True:
                r = cursor.fetchone()
                if not r:
                    break
                segbase = r[1] * SegmentSize.db_segment_size
                if r[2] == 1:
                    record_adder(recordsets[u],
                                 segments[u],
                                 (r[0], segbase + r[3]),
                                 record_number_to_game_number)
                    continue
                segment_cursor.execute(
                    ''.join(('select * from sgGame where rowid = ',
                             str(r[3]))))
                while True:
                    sr = segment_cursor.fetchone()
                    if not sr:
                        break
                    if len(sr[0]) == SegmentSize.db_segment_size_bytes:
                        for e, b in enumerate(sr[0]):
                            if not b:
                                continue
                            for eb, bitvalue in enumerate(
                                (128, 64, 32, 16, 8, 4, 2, 1)):
                                if bitvalue & b:
                                    record_adder(
                                        recordsets[u],
                                        segments[u],
                                        (r[0], segbase + eb + e * 8),
                                        record_number_to_game_number)
                    else:
                        rnl = sr[0]
                        for i in range(0, len(rnl), 2):
                            record_adder(
                                recordsets[u],
                                segments[u],
                                (r[0],
                                 segbase + int.from_bytes(rnl[i:i+2], 'big')),
                                record_number_to_game_number)
            del segment_cursor
    del cursor
    conn.close()


if __name__ == '__main__':
    runner = unittest.TextTestRunner
    loader = unittest.defaultTestLoader.loadTestsFromTestCase
    
    runner().run(loader(_DatabaseEncoders))

    # The Sqlite3base class tests.
    
    runner().run(loader(Sqlite3base_init))

    runner().run(loader(Sqlite3base_open_context))
    runner().run(loader(Sqlite3base_close_context))
    runner().run(loader(Sqlite3base_close_database))

    runner().run(loader(Sqlite3base_backout))
    runner().run(loader(Sqlite3base_close_contexts))
    runner().run(loader(Sqlite3base_commit))
    runner().run(loader(Sqlite3base_db_compatibility_hack))
    runner().run(loader(Sqlite3base_exists))
    runner().run(loader(Sqlite3base_files_exist))
    runner().run(loader(Sqlite3base_make_cursor))
    runner().run(loader(Sqlite3base_get_database_folder))
    runner().run(loader(Sqlite3base_get_database))
    runner().run(loader(Sqlite3base_get_database_instance))
    runner().run(loader(Sqlite3base_get_first_primary_key_for_index_key))
    runner().run(loader(Sqlite3base_get_primary_record))
    runner().run(loader(Sqlite3base_is_primary))
    runner().run(loader(Sqlite3base_is_primary_recno))
    runner().run(loader(Sqlite3base_is_recno))
    runner().run(loader(Sqlite3base_open_contexts))
    runner().run(loader(Sqlite3base_get_packed_key))
    runner().run(loader(Sqlite3base_decode_as_primary_key))
    runner().run(loader(Sqlite3base_encode_primary_key))
    runner().run(loader(Sqlite3base_use_deferred_update_process))
    runner().run(loader(Sqlite3base_initial_database_size))
    runner().run(loader(Sqlite3base_increase_database_size))
    runner().run(loader(Sqlite3base_do_database_task))
    runner().run(loader(Sqlite3base_make_recordset_key))
    runner().run(loader(Sqlite3base_make_recordset_key_startswith))
    runner().run(loader(Sqlite3base_make_recordset_key_range))
    runner().run(loader(Sqlite3base_make_recordset_all))
    runner().run(loader(Sqlite3base_file_records_under))
    
    runner().run(loader(Sqlite3base_method_classes))
    runner().run(loader(Sqlite3base_set_defer_update))
    runner().run(loader(Sqlite3base_unset_defer_update))
    runner().run(loader(Sqlite3base_do_deferred_updates))

    # The Sqlite3api class tests.  These take a total of about 6 minutes.

    runner().run(loader(Sqlite3api_init))

    runner().run(loader(Sqlite3api_put_instance))
    runner().run(loader(Sqlite3api_delete_instance))
    runner().run(loader(Sqlite3api_edit_instance))
    
    runner().run(loader(Sqlite3api_put_instance_with_holes))

    # At 17 March 2016, FreeBSD 10.1 with portsnap dated 7 February 2016, and
    # 17 August 2017, FreeBSD 10.3 with portsnap dated 8 April 2017, the test
    # Sqlite3api_delete_instance.test_delete_instance_05 usually failed
    # 'sqlite3.OperationalError: disk I/O error' at different points in
    # processing from run to run.  This test passed sometimes.
    # At 26 August 2017 the tests were changed so the apsw, sqlite3, bsddb3,
    # and dptdb, versions of the tests used the 'database_interface.py' common
    # code rather than separate copies.  It was immediately apparent that the
    # OperationalError exception was not happening any more, well at least for
    # three consecutive runs which had not been seen before.
    # I do not know why the change has this effect: but leave these comments
    # in case the problem comes back somehow.

    # Test Sqlite3api_put_instance_with_holes.test_holes_01 starts with the
    # code in Sqlite3api_delete_instance.test_delete_instance_05 but has never
    # suffered this error.

    # The equivalent tests for apsw do not suffer the error.
